# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from typing import List

sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------

project = "ape"
copyright = "2021, ApeWorX LTD"
author = "ApeWorX LTD"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_multiversion",
    "sphinx_rtd_theme",
    "autoapi.extension",
]
autoapi_type = "python"
autoapi_dirs = ["../src"]
autoapi_template_dir = "_autoapi_templates"
autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns: List[str] = ["_autoapi_templates/python/*.rst"]


def autoapi_skip_members(app, what, name, obj, skip, options):
    if "harambe" in name:
        return True
    return False


def setup(sphinx):
    sphinx.connect("autoapi-skip-member", autoapi_skip_members)


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


def fixpath(p):
    """
    Used to rewrite relative CSS paths to absolute paths in templates so we can
    get away with having only one copy of the CSS and web fonts.
    """
    new = f"/{project}/latest/_static/" + p.split("_static")[1]
    return new


html_context = {
    "fixpath": fixpath,
}
